import os
import shutil
import frag_pele
from peleffy.topology import Molecule, Topology, RotamerLibrary
from peleffy.forcefield import OpenForceField, OPLS2005ForceField
from peleffy.template import Impact
from peleffy.utils import get_data_file_path
import frag_pele.Covalent.correct_template_of_backbone_res as cov
import frag_pele.constants as c
from frag_pele.Helpers import folder_handler
from frag_pele.Helpers.plop_rot_temp import prepare_pdb
from peleffy.utils import Logger

logger = Logger()
logger.set_level('WARNING')

def create_template_path(path, name, forcefield='OPLS2005', protein=False, templates_generated=False):
    if templates_generated:
        templ_string = "templates_generated"
    else:
        templ_string = ""
    if forcefield == 'OPLS2005' and protein:
        path = os.path.join(path, 
                            'DataLocal/Templates/OPLS2005/Protein',
                            templ_string, name.lower())
    if forcefield == 'OFF' and protein:
        path = os.path.join(path, 
                            'DataLocal/Templates/OpenFF/Protein',
                            templ_string, aminoacid.lower())
    if forcefield == 'OPLS2005' and not protein:
        path = os.path.join(path,         
                            'DataLocal/Templates/OPLS2005/HeteroAtoms',
                            templ_string, name.lower()+"z")
    if forcefield == 'OFF' and not protein:
        path = os.path.join(path,
                            'DataLocal/Templates/OpenFF/Parsley',
                            templ_string, name.lower()+"z")
    return path

def get_template_and_rot(pdb, forcefield='OPLS2005', template_name='grw', aminoacid=False, outdir='.', rot_res=30,
                         contrained_atoms=None, aminoacid_type=None, sch_path=c.SCHRODINGER):
    p, pdb_name = os.path.split(pdb)
    out = pdb_name.split(".pdb")[0] + "_p" + ".pdb"
    currdir = os.getcwd()
    pdb_dir = os.path.dirname(pdb)
    # Check if the residue is an amino-acid from the library
    path = os.path.dirname(frag_pele.__file__)
    aa_pdb = os.path.join(path, f"Templates/Aminoacids/{aminoacid_type}.pdb")
    if not os.path.exists(aa_pdb) or aminoacid_type == None:
        output_pdb = os.path.join(pdb_dir,out)
        print(f"{aa_pdb} does not exist, using {output_pdb} instead")
    else:
        output_pdb = aa_pdb
    # Check if the output path exist to dont repeat calculations
    if not os.path.exists(output_pdb):
        os.chdir(pdb_dir)
        prepare_pdb(pdb_in=pdb_name, 
                    pdb_out=out, 
                    sch_path=sch_path)
        os.chdir(currdir)
    os.environ['SCHRODINGER'] = sch_path
    template_path = create_template_path(outdir, template_name, forcefield, aminoacid, True)
    if aminoacid:
        print("Aminoacid template")
        if not contrained_atoms:
            contraints = [' CA ', ' C  ', ' N  ']
        else:
            contraints = contrained_atoms
        m = Molecule(output_pdb, 
                     core_constraints=contraints,
                     rotamer_resolution=rot_res)
        print(f"Using PDB from {output_pdb} to generate the template!")
    else:
        print("Heteroatom template")
        if not contrained_atoms:
            m = Molecule(output_pdb,
                         rotamer_resolution=rot_res)
        else:
            m = Molecule(output_pdb,
                         core_constraints=contrained_atoms,
                         rotamer_resolution=rot_res)
    if forcefield == 'OPLS2005':
        ff = OPLS2005ForceField()
    if forcefield == 'OFF': # Not tested yet
        ff = OpenForceField('openff_unconstrained-1.2.0.offxml')
    parameters = ff.parameterize(m)
    topology = Topology(m, parameters)
    if forcefield == 'OFF': # Not tested yet
        from peleffy.solvent import OBC2
        solvent = OBC2(topology)
        solvent.to_file(os.path.join(outdir,
                                     "DataLocal/OBC/ligandParams.txt"))

    impact = Impact(topology)
    impact.to_file(template_path) 
    print("Template in {}.".format(template_path))
    rot_path = os.path.join(outdir, 
                            "DataLocal/LigandRotamerLibs/{}.rot.assign".format(template_name.upper()))
    rotamer_library = RotamerLibrary(m)
    rotamer_library.to_file(rot_path)
    print("Rotamer library stored in {}".format(rot_path))

def add_off_waters_to_datalocal(outdir):
    path = os.path.dirname(frag_pele.__file__)
    shutil.copy(os.path.join(path, "Templates/OFF/hohz"),
                os.path.join(outdir, "DataLocal/Templates/OpenFF/Parsley/hohz"))
 
def get_datalocal(pdb, outdir='.', forcefield='OPLS2005', template_name='grw', aminoacid=False, rot_res=30,
                  constrainted_atoms=None, aminoacid_type=None, sch_path=c.SCHRODINGER):
    folder_handler.check_and_create_DataLocal(working_dir=outdir)
    get_template_and_rot(pdb, forcefield=forcefield, template_name=template_name, 
                         aminoacid=aminoacid, outdir=outdir, rot_res=rot_res,
                         contrained_atoms=constrainted_atoms, aminoacid_type=aminoacid_type,
                         sch_path=sch_path)
#    if forcefield == 'OFF':
#        add_off_waters_to_datalocal(outdir)



import os
import frag_pele.Growing.template_fragmenter as tf

BACKBONE_ATOMS = ['_N__', '_C__', '_CA_']

def replace_atom_template(original, to_replace, atomname):
    original_indx = original.find_index_of_atom_name(atomname)
    atom_to_replace = original.list_of_atoms[original_indx]
    to_replace_indx = to_replace.find_index_of_atom_name(atomname) 
    to_replace.replace_atom(to_replace_indx, atom_to_replace)   

def replace_bond_template(original, to_replace, atomname):
    origin_atom_indx = original.find_index_of_atom_name(atomname)
    origin_bond_indx = original.find_bond_from_atom(origin_atom_indx)
    repl_atom_indx = to_replace.find_index_of_atom_name(atomname)
    repl_bond_indx = to_replace.find_bond_from_atom(repl_atom_indx)
    for o_indx in origin_bond_indx:
        bond_to_replace = original.list_of_bonds[o_indx]
        or_atom_1 = original.list_of_atoms[o_indx[0]].pdb_atom_name
        or_atom_2 = original.list_of_atoms[o_indx[1]].pdb_atom_name
        for r_indx in repl_bond_indx:
            rep_atom_1 = to_replace.list_of_atoms[r_indx[0]].pdb_atom_name
            rep_atom_2 = to_replace.list_of_atoms[r_indx[1]].pdb_atom_name
            if (or_atom_1 == rep_atom_1 and or_atom_2 == rep_atom_2) or (or_atom_1 == rep_atom_2 and or_atom_2 == rep_atom_1):
                print("Replacing bond between {} and {}. Indexes {}".format(or_atom_1, or_atom_2, r_indx))
                bond_to_replace.atom1 = r_indx[0] 
                bond_to_replace.atom2 = r_indx[1]
                to_replace.replace_bond(r_indx, bond_to_replace)


def replace_theta_template(original, to_replace, atomname):
    origin_atom_indx = original.find_index_of_atom_name(atomname)
    origin_theta_indx = original.find_theta_from_atom(origin_atom_indx)
    repl_atom_indx = to_replace.find_index_of_atom_name(atomname)
    repl_theta_indx = to_replace.find_theta_from_atom(repl_atom_indx)
    for o_indx in origin_theta_indx:
        theta_to_replace = original.list_of_thetas[o_indx]
        or_atom_1 = original.list_of_atoms[o_indx[0]].pdb_atom_name
        or_atom_2 = original.list_of_atoms[o_indx[1]].pdb_atom_name
        or_atom_3 = original.list_of_atoms[o_indx[2]].pdb_atom_name
        for r_indx in repl_theta_indx:
            rep_atom_1 = to_replace.list_of_atoms[r_indx[0]].pdb_atom_name
            rep_atom_2 = to_replace.list_of_atoms[r_indx[1]].pdb_atom_name
            rep_atom_3 = to_replace.list_of_atoms[r_indx[2]].pdb_atom_name
            if (or_atom_1 == rep_atom_1 and or_atom_2 == rep_atom_2 and or_atom_3 == rep_atom_3) or \
               (or_atom_1 == rep_atom_3 and or_atom_2 == rep_atom_2 and or_atom_3 == rep_atom_1):
                print("Replacing theta between {}, {} and {}. Indexes {}".format(or_atom_1, or_atom_2, or_atom_3, r_indx))
                theta_to_replace.atom1 = r_indx[0]
                theta_to_replace.atom2 = r_indx[1]
                theta_to_replace.atom3 = r_indx[2]
                to_replace.replace_theta(r_indx, theta_to_replace)


def correct_template(template, aminoacid_path="Data/Templates/OPLS2005/Protein/gly", work_dir="."):
    templ = tf.TemplateImpact(template)
    templ.erease_atom_from_template('_HN_')
    templ.erease_atom_from_template('_HXT')
    templ_to_copy = tf.TemplateImpact(os.path.join(work_dir, aminoacid_path))
    for name in BACKBONE_ATOMS:
        replace_atom_template(templ_to_copy, templ, name)
        replace_bond_template(templ_to_copy, templ, name)
        replace_theta_template(templ_to_copy, templ, name)
        templ.write_template_to_file(template)

def delete_atom_from_template(template, atomname):
    templ = tf.TemplateOPLS2005(template)
    if " " in atomname:
        atomname = atomname.replace(" ", "_")
    templ.erease_atom_from_template(atomname)
    templ.write_template_to_file(template)



import lib_prep.pdb_modifier as pdm

LIST_OF_IONS = ["ZN", "MN", "FE", "CO", "NI", "CA", "CD"]

class Corrector(pdm.PDB):
    def __init__(self, pdb_file, original_reschain, original_resnum, new_resname, ligand_chain=None):
        pdm.PDB.__init__(self, in_pdb=pdb_file)
        self.original_reschain = original_reschain
        self.original_resnum = original_resnum
        self.new_resname = new_resname
        self.ligand_chain = ligand_chain
        self.old_res_lines = self.get_residue(self.original_reschain,
                                              self.original_resnum)
        self.new_res_lines = self.get_atoms_of_resname(self.new_resname)
        self.remove_lines(self.new_res_lines)
        self.correct_atom_section_new_res_lines()
        self.correct_chain_and_resnum_new_res_lines()
        self.add_lines_at_position(self.new_res_lines, 
                                   self.find_index_of_line(self.old_res_lines[0]))
        self.remove_lines(self.old_res_lines)
        self.content = "".join(self.lines)


    def correct_atom_section_new_res_lines(self):
        correction = []
        for line in self.new_res_lines:
            if line.startswith("HETATM"):
                line = list(line)
                line[0:6] = "ATOM  "
                line = "".join(line)
            correction.append(line)
        self.new_res_lines = correction
    
    def check_and_rename_old_resname(self):
        correction = []
        for line in self.lines:
            resname = pdm.get_resname_from_line(line)
            reschain = pdm.get_chain_from_line(line)
            resnum = pdm.get_resnum_from_line(line)
            if str(reschain).strip() == str(self.original_reschain).strip() and \
               str(resnum).strip() == str(self.original_resnum).strip() and \
               str(resname).strip() == str(self.new_resname).strip():
                line = pdm.set_resname_to_line(line, "RES")
            correction.append(line)
        self.lines = correction
 
    def clear_heteroatoms(self):
        to_keep = []
        for line in self.lines:
            if line.startswith("HETATM"):
                if pdm.get_chain_from_line(line) != self.ligand_chain:
                    continue
                if pdm.get_resname_from_line(line) != "HOH":
                    continue
                if pdb.get_resname_from_line not in LIST_OF_IONS:
                    continue
                to_keep.append(line)
        self.lines = to_keep

    def correct_chain_and_resnum_new_res_lines(self):
        correction = []
        for line in self.new_res_lines:
            line = pdm.set_chain_to_line(line, self.original_reschain)
            line = pdm.set_resnum_to_line(line, self.original_resnum)
            correction.append(line)
        self.new_res_lines = correction

    def get_indexes_from_list_of_lines(self, list_of_lines):
        indexes = []
        for line in list_of_lines:
            idx = pdm.get_atom_index_from_line(line)
            indexes.append(idx)
        return indexes

    def add_lines_at_position(self, lines, position):
        lines.reverse()
        for line in lines:
            self.lines.insert(position, line)
   
    def find_index_of_line(self, line):
        index = self.lines.index(line)
        return index

    def remove_lines(self, lines):
        for line in lines:
            self.lines.remove(line)
        

def correct_pdb(pdb_file, reschain, resnum, new_resname):
    correct = Corrector(pdb_file, reschain, resnum, new_resname)
    correct.overwrite_pdb()

def delete_atom_ligand(pdb_file, ligchain, atomname):
    if len(atomname) != 4:
        raise ValueError("The atom name of the linking atom must be a string of length 4! Current length: {}".format(len(atomname)))
    pdb = pdm.PDB(pdb_file)
    atomlines = pdb.atom_section.split("\n\n")
    for l in atomlines:
        at_name = l[12:16]
        at_chain = l[21:22]
        if at_name == atomname and at_chain == ligchain:
            print("Deleting {} from {}".format(atomname, ligchain))
            for n, line in enumerate(pdb.lines):
                if l in line:
                    del pdb.lines[n]
            pdb.content = "".join(pdb.lines)
            pdb.overwrite_pdb()



import os
import time
import subprocess
from frag_pele.Helpers import correct_fragment_names
import frag_pele.constants as c


class CovalentPDB:

    content = None
    ligand = None
    residue_type_linked = None
    conect_str = []
    conects = []

    def __init__(self, pdb_file, lig_chain="L", res_num=None, 
                 res_chain=None, lig_link=None, res_link=None):
        self.pdb_file = pdb_file
        self.lig_chain = lig_chain
        self.res_num = res_num
        self.res_chain = res_chain
        self.lig_link = lig_link
        self.res_link = res_link
        self._read_pdb_content()
        if not self._check_if_exists_conects() and (res_num==None or res_chain==None):
            raise ValueError("PDB file does not contain CONECTs." 
                             "Please, fill the number and chain of the residue covalently linked to the ligand")
        self._read_ligand()
        self._read_conects()
        self._get_residue_info_linked()
    
    def _read_pdb_content(self):
        with open(self.pdb_file, "r") as infile:
            self.content = infile.read()    
 
    def _check_if_exists_conects(self):
        pdb_lines = self.content.split("\n")
        for l in pdb_lines:
            if l.startswith("CONECT"):
                return True
        return False

    def _read_ligand(self):
        lig_lines = []
        for l in self.content.split("\n"):
            if l.startswith("HETATM"):
                if self.lig_chain == l[21]:
                    lig_lines.append(l)
        self.ligand = "\n".join(lig_lines)

    def _read_conects(self):
        for l in self.content.split("\n"):
            if l.startswith("CONECT"):
                group = l.split("CONECT")[1].split()
                group = list(map(int, group))
                self.conects.append(group)
    
    def _read_conects_str(self):
        lines = []
        for l in self.content.split("\n"):
            if l.startswith("CONECT"):
                lines.append(l)
        return "\n".join(lines)

    def _get_ligand_residue_connection(self):
        conections = []
        all_indexes = get_indexes(self.content)
        ligand_indexes = get_indexes(self.ligand)
        residue_indexes = list(set(all_indexes) - set(ligand_indexes))
        if self.conects:
            for c_group in self.conects:
                # Check if any conection group contains atoms of the ligand and residues at the same time
                if set(c_group).intersection(set(ligand_indexes)) and set(c_group).intersection(set(residue_indexes)):
                    conections.append(c_group)
        return conections
   
    def _get_residue_info_linked(self):
        conections = self._get_ligand_residue_connection()[0]
        for line in self.content.split("\n"):
            if line.startswith("HETATM") and line[6:11].strip() == str(conections[0]):
                self.lig_link = line[12:15].strip()
            if line.startswith("ATOM") and line[6:11].strip() == str(conections[-1]):
                self.res_num = line[23:26].strip()
                self.res_chain = line[21]
                self.res_link = line[12:15].strip()
                self.residue_type_linked = line[17:20].strip()
                print(line[17:20].strip())

    def get_residue_attached(self):
        lines = []
        for l in self.content.split("\n"):
            if l.startswith("ATOM"):
                if l[23:26].strip() == self.res_num and l[21] == self.res_chain:
                    lines.append(l)
        return "\n".join(lines)
  
    def replace_residue_name(self, new_name):
        lines = []
        for l in self.content.split("\n"):
            if l.startswith("ATOM"):
                if l[23:26].strip() == self.res_num and l[21] == self.res_chain:
                    l = list(l)
                    l[17:20] = new_name
                    l = "".join(l)
            lines.append(l)
        self.content = "\n".join(lines)

    def get_ligand_and_res(self):
        lines = []
        mix = self.get_residue_attached() + "\n" + self.ligand
        for l in mix.split("\n"):
            l = list(l)
            l[0:6] = "HETATM"
            l[17:20] = "MIX"
            l[21:22] = "L"
            l = "".join(l)
            lines.append(l)
        mix = "\n".join(lines)
        con = self._read_conects_str()
        return mix + "\n" + con

    def get_pdb_prepared(self):
        restype = self.residue_type_linked
        new_resname = restype[0:2] + "Y"
        self.replace_residue_name(new_resname) 
        

def get_indexes(pdb_fragment):
    indexes = []
    for l in pdb_fragment.split("\n"):
        if l.startswith("HETATM") or l.startswith("ATOM"):
            indexes.append(int(l[7:11].strip()))
    return indexes

def write_string_to_file(string, filename):
    with open(filename, "w") as outf:
        outf.write(string)

def read_file(filename):
    with open(filename) as inf:
        cont = inf.read()
    return cont

def prepare_pdb(pdb_in, pdb_out, sch_path):
    command = [os.path.join(sch_path, "utilities/prepwizard"), pdb_in, pdb_out, "-noepik", "-noprotassign",
              "-noccd", "-noimpref"]
    subprocess.call(command)

def check_and_solve_duplicate_atomnames(pdb_file, chain="L"):
    content = read_file(pdb_file)
    elements = []
    pdb_out = []
    names = []
    new_names = []
    for line in content.split("\n"):
        if line[0:6] == "HETATM" and line[21:22] == chain:
            element = line[76:78]
            elements.append(element)
            atom_names = line[12:16]
            names.append(atom_names)
            list_of_unique_elements = set(elements)
            counters = {i: elements.count(i) for i in list_of_unique_elements}
            counter = counters[line[76:78]]
            set_to_check = set(names)
            list_to_check = sorted(list(set_to_check))
            sorted_list_names = sorted(names)
            if list_to_check != sorted_list_names:
                new_name = "{}{}".format(line[76:78].strip().upper(), counter)
                while True:
                    new_name = new_name + " "
                    if len(new_name) == 4:
                        break
                line = list(line)
                line[12:16] = new_name
                line = "".join(line)
                if len(new_name) > 4:
                    raise ValueError("Length of the string {} is too long. Only 4 characters accepted.".format(new_name))
                new_names.append(new_name)
        pdb_out.append(line)
    write_string_to_file("\n".join(pdb_out), pdb_file)
    return new_names

def create_template(pdb_file, sch_python=c.SCHRODINGER_PY_PATH, 
                    plop_script_path="../PlopRotTemp_S_2017/ligand_prep.py", rotamers="10.0", 
                    out_templates_path=".", path_to_lib="."):
    cmd = "{} {} {} {} {} {}".format(sch_python, plop_script_path, pdb_file, rotamers,
                                     out_templates_path, path_to_lib)
    print(cmd)
    try:
        subprocess.call(cmd.split())
    except OSError:
        raise OSError("Path {} not foud. Change schrodinger path under frag_pele/constants.py".format(sch_python))

def prepare_complex_and_extract_files_for_covalent(pdb_file, sch_path="/gpfs/projects/bsc72/SCHRODINGER_ACADEMIC",
                                                  lig_chain="L", res_num=None, res_chain=None, lig_link=None, 
                                                  res_link=None, rotamers="10.0", out_templates_path=".", 
                                                  path_to_lib="."):
    pdb = CovalentPDB(pdb_file, lig_chain, res_num, res_chain, lig_link, res_link)
    # Modifying PDB content
    pdb.get_pdb_prepared()
    mix = pdb.get_ligand_and_res()
    res = pdb.get_residue_attached()
    # Creating LIGAND, RESIDUE, MIX(COMBINATION) and COMPLEX
    write_string_to_file(pdb.ligand, "LIG.pdb")
    write_string_to_file(res, "{}.pdb".format(pdb.residue_type_linked))
    write_string_to_file(mix, "MIX.pdb")
    write_string_to_file(pdb.content, pdb_file)
    # Preparing them with Schrodinger (Correcting free atoms)
    prepare_pdb("LIG.pdb", "LIG_h.pdb", sch_path=sch_path)
    prepare_pdb("MIX.pdb", "MIX_h.pdb", sch_path=sch_path)
    prepare_pdb("{}.pdb".format(pdb.residue_type_linked), 
                "{}_h.pdb".format(pdb.residue_type_linked), 
                sch_path=sch_path)
    time.sleep(3) # Waiting queue system
    # Solving duplicate names and getting the new H atoms
    h_to_del = check_and_solve_duplicate_atomnames("LIG_h.pdb", chain=lig_chain)
    check_and_solve_duplicate_atomnames("MIX_h.pdb", chain=lig_chain)
    # Creating templates of LIGAND and COMBINATION
    create_template("LIG_h.pdb", rotamers=rotamers, out_templates_path=out_templates_path, 
                    path_to_lib=path_to_lib)
    create_template("MIX_h.pdb", rotamers=rotamers, out_templates_path=out_templates_path, 
                    path_to_lib=path_to_lib)
    print("Files prepared successfully!")
    return pdb.residue_type_linked, pdb.res_link, pdb.lig_link, h_to_del


#def prepare_templates_for_covalent():


def prepare_covalent_growing(pdb_file, sch_path="/gpfs/projects/bsc72/SCHRODINGER_ACADEMIC",
                             lig_chain="L", res_num=None, res_chain=None, lig_link=None,
                             res_link=None, rotamers="10.0", out_templates_path=".",
                             path_to_lib="."):
    prepare_complex_and_extract_files_for_covalent("covdock_sar26lu7_14.pdb")




import os
import glob
import numpy as np
import networkx as nx
import peleffy
from frag_pele.constants import SCHRODINGER
from peleffy.topology import Topology
from peleffy.forcefield import OpenForceField, OPLS2005ForceField
from peleffy.topology import Molecule
from peleffy.utils import Logger
from peleffy.topology import molecule
from peleffy.utils.toolkits import RDKitToolkitWrapper

class ComputeDihedrals(object):
    """
    A class to produce a library of dihedral angles.
    """
    def __init__(self, pdb_file, forcefield='OPLS2005'):
        """
        Initializes a ComputeDihedrals object
        Parameters
        ----------
        topology_obj : An peleffy.topology.Topology
            A Topology object that contains the ligand's information
        mode: str
            Whether to extract all dihedrals or only those marked as flexible
        """
        self._pdb_file = pdb_file
        self._forcefield = forcefield
        self._topology, self._molecule = self.load_molecule()
        self.dihedral_library = {}

    def load_molecule(self):
        os.environ['SCHRODINGER'] = SCHRODINGER 
        m = Molecule(self._pdb_file)
        if self._forcefield == 'OPLS2005':
            ff = OPLS2005ForceField()      
        elif self._forcefield == 'OpenForceField':
            ff = OpenForceField('openff_unconstrained-1.2.0.offxml')
        else:
            raise ValueError("Not valid ForceField. Pick between 'OPLS2005' and 'OpenForceField'")
        parameters = ff.parameterize(m)                                        
        top = peleffy.topology.Topology(m, parameters) 
        return top, m

    def list_all_dihedrals(self):
        dihedrals = []
        seen_dihedrals = set()
        for proper in self._topology.propers:
            dihedral_indexes = (proper.atom1_idx, proper.atom2_idx, proper.atom3_idx, proper.atom4_idx)
            if dihedral_indexes in seen_dihedrals:
                continue
            seen_dihedrals.add(dihedral_indexes)
            dihedrals.append(list(dihedral_indexes))
        return dihedrals

    def calculate_cluster_angles(self, dihedral_list):
        """
        Calculate dihedral angles from pdb
        Parameters
        ----------
        pdb_file: str
            Path to the cluster representative conformation
        dihedral_list: list
            List of the tuples containing the atoms that form the dihedrals
        match_indexes: bool
            Whether to use the atom indices from the dihedral list or match to
            the cluster structure before
        """
        rdkit_wrapper = RDKitToolkitWrapper()
        pdb_dihedrals = []
        # use the input molecule as template since the cluster structures
        # probably will not have proper stereochemistry
        mol = molecule.Molecule(self._pdb_file, connectivity_template=self._molecule.rdkit_molecule)
        # we use substructure matching to ensure that the indices in the
        # clusters pdb and the input ligand are the same
        for dihedral in dihedral_list:
            names = [self._topology.atoms[atom].PDB_name for atom in dihedral]
            angle = get_dihedral(mol, *dihedral, units="degrees")
        
            pdb_dihedrals.append(names+[angle])
        self.dihedral_library[self._pdb_file] = pdb_dihedrals
    
    def _calculate_all_dihedrals(self):
        dihedrals = self.list_all_dihedrals()
        self.calculate_cluster_angles(dihedrals)


    def calculate(self):
        """
        Calculate dihedrals library from the bce output
        """
        logger = Logger()
        logger.info(' - Calculating dihedral library')
        self._calculate_all_dihedrals()

def get_dihedral(mol, atom1, atom2, atom3, atom4, units="radians"):
    """
    It calculates the value of the dihedral angle in the specified units
        (default radians)
    Parameters
    ----------
    molecule : an offpele.topology.Molecule
        The offpele's Molecule object
    atom1 : int
        Index of the first atom in the dihedral
    atom2 : int
        Index of the second atom in the dihedral
    atom3 : int
        Index of the third atom in the dihedral
    atom4 : int
        Index of the fourth atom in the dihedral
    units : str
        The units in which to calculate the angle (default is radians, can
        be radians or degrees)
    """
    from rdkit.Chem import rdMolTransforms
    if units == "degrees":
        angle = rdMolTransforms.GetDihedralDeg(mol.rdkit_molecule.GetConformer(), atom1, atom2, atom3, atom4)
    else:
        angle = rdMolTransforms.GetDihedralRad(mol.rdkit_molecule.GetConformer(), atom1, atom2, atom3, atom4)
    return angle


def select_dihedrals(input_dihedrals_list, selected_dihedrals_atoms):
    selected_dihedrals = []
    for dihedrals in input_dihedrals_list:
        atom1, atom2, atom3, atom4, angle = dihedrals
        atoms = [atom1, atom2, atom3, atom4]
        for atoms_of_dih in selected_dihedrals_atoms:
            if atoms_of_dih == atoms or atoms_of_dih[::-1] == atoms: # Try in both reading senses
                print(f"Constraining dihedral: {atoms}")
                selected_dihedrals.append([*atoms, angle])
    if len(selected_dihedrals) == 0:
        raise ValueError(f"The dihedral formed by {selected_dihedrals_atoms} were not found in ligand dihedrals.")
    return selected_dihedrals




import glob
import os
import argparse
import joblib
import mdtraj as md
import pandas as pd

def parseArguments():
    """
        Parse the command-line options
        :returns: str, int, int --  path to file to results folder,
            index of the first atom,
            index of the second atom
    """
    desc = "It includes the atom-atom distance of the specified ones to report files\n"
    parser = argparse.ArgumentParser(description=desc)
    required_named = parser.add_argument_group('required named arguments')
    required_named.add_argument("sim_folder", type=str, help="Path to the simulation results.")

    required_named.add_argument("-a", "--atoms", type=int, nargs='+', action='append', 
                                help="List of pairs of indexes with atoms to compute the distance between them.")
    parser.add_argument("-t", "--traj", default="trajectory_",
                        help="Trajectory file prefix.")
    parser.add_argument("-r", "--rep", default="report_",
                        help="Report file prefix.")
    parser.add_argument("-p", "--proc", type=int, default=4,
                        help="Number of processors to paralellize the computation.")
    args = parser.parse_args()
    return args.sim_folder, args.atoms, args.traj, args.rep, args.proc


def compute_atom_atom_dist(infile, atoms_list):
    distances = []
    names = []
    traj = md.load_pdb(infile)
    print(atoms_list)
    for at_pair in atoms_list:
        name ="{}-{}".format(traj.topology.atom(at_pair[0]), traj.topology.atom(at_pair[1]))
        distance = md.compute_distances(traj, [at_pair])
        distances.append(distance)
        names.append(name)
    return distances, names


def compute_distances_from_report(atomlist, report, trajectory):
    distances, colnames = compute_atom_atom_dist(trajectory, atomlist)
    new_lines = []
    with open(report) as rep:
        rep_lines = rep.readlines()
        rep_lines = [x.strip("\n") for x in rep_lines]
        for ind, line in enumerate(rep_lines):
            new_content = list(line.split("    "))
            if new_content[-1] == '':
                new_content = new_content[:-1]
            if ind == 0:
                for colname in colnames:
                    new_content.append(colname)
            else:
                for dist in distances:
                    value = "{:.3f}".format(dist[ind-1][0]*10)
                    new_content.append(value)
            new_line = "    ".join(new_content)
            new_lines.append(new_line)
    new_report = "\n".join(new_lines)
    new_rep_name = report.split("/")
    new_rep_name[-1] = "dist" + new_rep_name[-1]
    new_rep_name = "/".join(new_rep_name)
    with open(new_rep_name, "w") as out:
        out.write(new_report)
    print("{} completed".format(new_rep_name))

def compute_simulation_distance(sim_folder, atomlist, traj_pref="trajectory_", report_pref="report_", processors=4):
    trajectories = sorted(glob.glob("{}*".format(os.path.join(sim_folder, traj_pref))))
    reports = sorted(glob.glob("{}*".format(os.path.join(sim_folder, report_pref))))
    joblib.Parallel(n_jobs=processors)(joblib.delayed(compute_distances_from_report)(atomlist, report, traj) for report, traj in zip(reports, trajectories))

if __name__ == '__main__':
    sim_fold, atom_list, traj, report, processors = parseArguments()
    compute_simulation_distance(sim_fold, atom_list, traj, report, processors)



import sys
import os
import socket

DIR = os.path.dirname(__file__)

##############################################

# PUBLIC CONSTANTS (to change by the user)
# Preparation inputs to grow

machine = socket.getfqdn()
print(machine)

# Paths definitions (IMPORTANT!)
if "bsc.mn" in machine:
    # PELE parameters
    PATH_TO_PELE = "/gpfs/projects/bsc72/PELE++/mniv/V1.7.1/bin/PELE-1.7.1_mpi"
    PATH_TO_PELE_DATA = "/gpfs/projects/bsc72/PELE++/mniv/V1.7.1/Data"
    PATH_TO_PELE_DOCUMENTS = None # Data and documents is added automatically from PELE >= 1.6
    PATH_TO_LICENSE = "/gpfs/projects/bsc72/PELE++/license"
    # PlopRotTemp parameters
    SCHRODINGER = "/gpfs/projects/bsc72/SCHRODINGER_ACADEMIC"
elif "bsccv" in machine:
    # PELE parameters
    PATH_TO_PELE = "/data/EAPM/PELE/PELE++/bin/rev12360/Pele_rev12360_mpi"
    PATH_TO_PELE_DATA = "/data/EAPM/PELE/PELE++/data/rev12360/Data"
    PATH_TO_PELE_DOCUMENTS = "/data/EAPM/PELE/PELE++/Documents/rev12360"
    PATH_TO_LICENSE = "/data/EAPM/PELE/PELE++/license"
    SCHRODINGER = "/data2/bsc72/SCHRODINGER_ACADEMIC"
else:
    # PELE parameters
    PATH_TO_PELE = "/home/user/pelepath/PELE_mpi"
    PATH_TO_PELE_DATA = None
    PATH_TO_PELE_DOCUMENTS = None
    PATH_TO_LICENSE = "/home/user/pelepath/licenses"
    SCHRODINGER = "/home/user/schrodingerVVVV-VV"

# FragPELE configuration
CONTROL_TEMPLATE = os.path.join(DIR, "Templates/control_template.conf")
GROWING_STEPS = 10
SELECTION_CRITERIA = "Binding Energy"
CPUS = 48
PELE_EQ_STEPS = 20  # PELE steps in the equilibration
RESTART = False
STEPS = 6  # PELE steps for growing step
BANNED_DIHEDRALS_ATOMS = None
BANNED_ANGLE_THRESHOLD = None

# PELE control file configuration
REPORT_NAME = "report"
TRAJECTORY_NAME = "trajectory"
RESULTS_FOLDER = "growing_output"
TEMPERATURE = 1000
MAX_OVERLAP = 0.70
MIN_OVERLAP = 0.50
SEED = 1279183
STEERING = 0
TRANSLATION_HIGH = 0.05 #Amstrongs
ROTATION_HIGH = 0.10 # Radiants
TRANSLATION_LOW = 0.02
ROTATION_LOW = 0.05 
RADIUS_BOX = 4  # Amstrongs
# PlopRotTemp configuration
ROTRES = 30

# Clustering parameters
DISTANCE_COUNTER = 4
CONTACT_THRESHOLD = 0.3
EPSILON = 0.5
CONDITION = "min"   #   min or max
METRICS_WEIGHTS = "linear"
NUM_CLUSTERS = 5
##############################################

# PRIVATE CONSTANTS (not to change)
PRE_WORKING_DIR = "pregrow"
TEMPLATES_PATH = "DataLocal/Templates/OPLS2005/HeteroAtoms/"
ROTAMERS_PATH = "DataLocal/LigandRotamerLibs/"
PDBS_OUTPUT_FOLDER = "clustering_PDBs"
OUTPUT_FOLDER = "growing_steps/"
TEMPLATES_FOLDER = "growing_templates"
CONFIG_PATH = "log_configure.ini"
PLOP_PATH = "PlopRotTemp_S_2017/ligand_prep.py"

# Messages constants
TEMPLATE_MESSAGE = "We are going to transform the template _{}_ into _{}_ in _{}_ steps! Starting..."
LINES_MESSAGE = "\n•*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.••*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.•*´¨`*•.¸¸.•\n"
SELECTED_MESSAGE = "\n============ Files selected ============\nControl file: {}\nPDB file: {}\nResults folder name: {}\nStep: {}\n"
FINISH_SIM_MESSAGE = "SIMULATION {} COMPLETED!!! "

# LISTS
AA_LIST = ["ALA", "ARG", "ASP", "CYS", "GLN", "GLY", "GLU", "HID", "HIE", "HIP", "HIS", "ILE", "LEU", "LYS", 
           "MET", "PHE", "PRO", "TRP", "TYR", "THR", "VAL"]
##############################################



import numpy as np
class RotamerModifier():
    def __init__(self, input_file):
        self.input_file = input_file
        self.content = None
        self.lines = []
        self.lines_splitted = []
        self.read_file()
        self.read_rotamers()

    def read_file(self):
        with open(self.input_file) as inp:
            self.lines = inp.readlines()
   
    def read_rotamers(self):
        for line in self.lines:
            split = line.split(" ")
            self.lines_splitted.append(split)

    def find_rotamer(self, atom1, atom2):
        indexes = []
        for n, line in enumerate(self.lines_splitted):
            if line[3] == "sidelib":
                if (line[5] == atom1 and line[6] == atom2) or \
                   (line[5] == atom2 and line[6] == atom1):
                    indexes.append(n)
        return indexes # The index of the list

    def delete_rotamer(self, atom1, atom2):
        indexes_to_del = self.find_rotamer(atom1, atom2)
        for index_to_del in indexes_to_del:
            print("Rotamer '{}' will be deleted".format(" ".join(self.lines_splitted[index_to_del]).strip("\n")))
        np.delete(self.lines_splitted, indexes_to_del)
        self.update_content()

    def update_content(self):
        self.lines = []
        for line in self.lines_splitted:
            self.lines.append(" ".join(line))
        self.content = "".join(self.lines)

    def overwrite_file(self):
        with open(self.input_file, "w") as out:
            out.write(self.content)

def delete_atoms_from_rot_lib(input_file, list_of_atom_pairs):
    rot = RotamerModifier(input_file)
    for pair in list_of_atom_pairs:
        rot.delete_rotamer(pair[0], pair[1])
    rot.overwrite_file()

def delete_atoms_from_rot_lib_by_list(input_file, list_of_atoms):
    rot = RotamerModifier(input_file)
    rot.read_rotamers()
    rotamers = rot.lines_splitted
    for r in rotamers:
        if r[3] == "sidelib":
            if r[5] in list_of_atoms and r[6] in list_of_atoms:
                rot.delete_rotamer(r[5], r[6])
    rot.overwrite_file()



import os
import time
import shutil
import subprocess
import frag_pele.constants as c

DATA_PATH = "Data/Templates/OPLS2005/Protein"

def create_template(pdb_file, sch_path=c.SCHRODINGER,
                    plop_script_path="../PlopRotTemp_S_2017/ligand_prep.py", rotamers="10.0",
                    out_templates_path=".", path_to_lib=".", cov_res=None, work_dir="."):
    if "/" in pdb_file:
        filename = pdb_file.split("/")[-1]
        filename = filename.split(".pdb")[0].lower()
    else:
        filename = pdb_file.split(".pdb")[0].lower()

    if cov_res:
       if os.path.exists(os.path.join(work_dir, DATA_PATH, filename)):
           print("Template found in {}".format(os.path.join(DATA_PATH, filename)))
           shutil.copy(os.path.join(work_dir, DATA_PATH, filename), 
                       os.path.join(work_dir, "DataLocal/Templates/OPLS2005/Protein/templates_generated"))
           return filename
       else:
           schrodinger_path = sch_path
           sch_python = os.path.join(sch_path, "utilities/python")
           prepare_pdb(pdb_file, "GRW.pdb", schrodinger_path)
           cmd = "{} {} {} {} {} {}".format(sch_python, plop_script_path, "GRW.pdb", rotamers,
                                            os.path.join(work_dir, 
                                            "DataLocal/Templates/OPLS2005/Protein/templates_generated"), 
                                            path_to_lib)

           print("Running PlopRotTemp...")
           print(cmd)
           try:
              subprocess.call(cmd.split())
           except OSError:
               raise OSError("Path {} not foud. Change schrodinger path under frag_pele/constants.py".format(sch_python))
    else:
        cmd = "{} {} {} {} {} {}".format(sch_python, plop_script_path, pdb_file, rotamers,
                                            out_templates_path, path_to_lib)
        print("Running PlopRotTemp...")
        print(cmd)
        try:
            subprocess.call(cmd.split())
        except OSError:
            raise OSError("Path {} not foud. Change schrodinger path under frag_pele/constants.py".format(sch_python))
    return filename    

def prepare_pdb(pdb_in, pdb_out, sch_path):
    command = [os.path.join(sch_path, "utilities/prepwizard"), pdb_in, pdb_out, "-noepik", "-noprotassign",
               "-noimpref", "-noccd", "-NOJOBID"]
    print(command)
    subprocess.call(command)



import os, argparse
from lib_prep.pdb_modifier import PDB


def parseArguments():
    """
        Parse the command-line options
    """
    desc = "It process PDB files putting the ligand part of a covalent ligand into the residue that it is attached with."
    parser = argparse.ArgumentParser(description=desc)
    required_named = parser.add_argument_group('required named arguments')
    required_named.add_argument("-i", "--input_pdb", type=str,  
                                help="Path to the input PDB complex.")
    required_named.add_argument("-rc", "--residue_chain", type=str,   
                                help="Chain of the residue that has the ligand covalently attached")
    required_named.add_argument("-rn", "--residue_number", type=str,
                                help="Residue number of the residue that has the ligand covalently attached")
    required_named.add_argument("-o", "--output_pdb", type=str,
                                help="Output PDB path.")
    parser.add_argument("-lr", "--ligand_resname", default="UNK",
                        help="Residue name of the ligand that is covalently attached to the previous residue")
    parser.add_argument("-lc", "--ligand_chain", default=" ",
                        help="Chain of the ligand that is covalently attached to the previous residue")
    parser.add_argument("-v", "--verbose", default=False, action='store_true',
                        help="Set true to activate all unneded prints.")
    parser.add_argument("-e", "--extract_ligand", default=True, action='store_false',
                        help="Set true to inactivate the extraction of a ligand in a PDB file.")
    args = parser.parse_args()

    return args.input_pdb, args.residue_chain, args.residue_number, args.output_pdb, \
           args.ligand_resname, args.ligand_chain, args.verbose, args.extract_ligand


class CovCorrector:
    def __init__(self, input_pdb, residue_chain, residue_number, ligand_resname='UNK', ligand_chain=' ',
                 verbose=False):
        """
        Class to correct PDB files with covalent ligands from Maestro's outputs.

        Parameters
        ----------
        input_pdb : str
            path to input PDB file
        residue_chain : str
            chain of the residue that has the ligand covalently attached
        residue_number : int
            number of the residue that has the ligand covalently attached
        ligand_resname : str
            residue name of the ligand that is covalently attached to the previous residue
        ligand_chain : str
            chain of the ligand that is covalently attached to the previous residue
        verbose : bool
            if true it will activate all the prints
        ----------

        """
        self._input_pdb = input_pdb
        self._residue_chain = residue_chain
        self._residue_number = residue_number
        if isinstance(self._residue_number, str):
            self._residue_number = int(self._residue_number)
        self._ligand_resname = ligand_resname
        self._ligand_chain = ligand_chain
        self._verbose = verbose
        self._pdb = PDB(in_pdb=self._input_pdb, chain=self._ligand_chain, resname=self._ligand_resname)
        self._ligand_lines = None
        self._ligand = None
        self._residue = None
        self._residue_type = None
        self._get_ligand_section()
        self._get_residue_info()

    def _get_ligand_section(self):
        """
        It gets and prints the lines of the PDB file which contains the assigned ligand_resname.
        It will fill the attribute self._ligand_lines and self._ligand
        """
        self._ligand_lines = self._pdb.get_atoms_of_resname(self._ligand_resname)
        self._ligand = ''.join(self._ligand_lines)
        if self._verbose:
            print('Ligand lines: \n' + self._ligand)

    def _get_residue_info(self):
        """
        It obtains the residue type of the residue that has the ligand covalently attached and fills
        the variable self.residue_type
        """
        residue_lines = self._pdb.get_residue(self._residue_chain, str(self._residue_number))
        self.residue = ''.join(residue_lines)
        self.residue_type = ''.join(residue_lines[0][17:20])
        if self._verbose:
            print('Residue lines: \n' + self.residue)
            print('Residue type: {}'.format(self.residue_type))

    def _correct_ligand_to_be_residue(self):
        """
        It corrects the ligand lines to transform it in to residue lines
        """
        new_ligand_lines = []
        for lig_line in self._ligand_lines:
            lig_line = list(lig_line)
            lig_line[0:6] = "ATOM  "  # Atom section
            lig_line[17:20] = self.residue_type  # Residue name
            lig_line[21:22] = self._residue_chain  # Chain
            lig_line[22:26] = "{:>4}".format(self._residue_number)  # Residue number, must be right-aligned
            new_ligand_lines.append(''.join(lig_line))
        self._ligand = ''.join(new_ligand_lines)
        self._ligand_lines = new_ligand_lines
        if self._verbose:
            print('New ligand lines:\n' + self._ligand)

    def _join_ligand_and_residue(self, reindexing=True):
        """
        It appends ligand lines onto the residue, and it can reindex atom ids.

        Parameters
        ----------
        reindexing : bool
            if it is true, it will reindex the atom IDs of the joining result
            
        Returns
        -------
        joining_result : list
            lines of the amino-acid with the ligand attached, all in the same residue
        """
        counter = 1
        joining_result = []
        for res_line in self.residue.split('\n')[0:-1]:  # The last line is empty because of the splitting
            res_line = list(res_line)
            if reindexing:
                res_line = list(res_line)
                res_line[6:11] = "{:>5}".format(counter)
                counter = counter + 1
            res_line[17:20] = self._ligand_resname
            res_line = ''.join(res_line)
            joining_result.append(res_line+'\n')
        for lig_line in self._ligand_lines:
            lig_line = list(lig_line)
            if reindexing:
                lig_line = list(lig_line)
                lig_line[6:11] = "{:>5}".format(counter)
                counter = counter + 1
            lig_line[17:20] = self._ligand_resname
            lig_line = ''.join(lig_line)
            joining_result.append(lig_line)
        return joining_result

    def correct(self, reindexing=True):
        """
        It joins the ligand and the amino-acid of a covalent ligand into a single residue, and both are moved to the
        residue position of the PDB file. This makes it compatible with PELE.

        Parameters
        ----------
        reindexing : bool
            if it is true, it will reindex the atom IDs of the joining result
        """
        global residue_idx
        new_pdb_lines = []
        self._correct_ligand_to_be_residue()
        residue_corrected = self._join_ligand_and_residue(reindexing)
        index_filled = False
        for n, line in enumerate(self._pdb.lines):
            if line.startswith('ATOM') and \
                    line[21:22] == self._residue_chain and \
                    int(line[22:26].strip()) == self._residue_number:
                if not index_filled:
                    residue_idx = n
                    index_filled = True
                continue
            if line.startswith('HETATM') and \
                    line[21:22] == self._ligand_chain and \
                    line[17:20].strip() == self._ligand_resname:
                continue
            new_pdb_lines.append(line)
        new_pdb_lines[residue_idx:residue_idx] = residue_corrected  # Insert the residue in the correct index
        self._pdb.lines = new_pdb_lines
        self._pdb.content = ''.join(new_pdb_lines)

    def write_file(self, output_file, extract_ligand=True):
        """
        It writes the content into an output PDB file.

        Parameters
        ----------
        output_file : str
            path of the output PDB file
        extract_ligand : bool
            if true it extracts the ligand in a separated PDB file
        """
        with open(output_file, "w") as out_pdb:
            out_pdb.write(self._pdb.content)
            print("PDB saved in {}.".format(output_file))
        if extract_ligand:
            lig_lines = self._pdb.get_atoms_of_resname(f"{self._ligand_resname}")
            ligand_block = "".join(lig_lines)
            out_path = os.path.dirname(output_file)
            with open(os.path.join(out_path, f"{self._ligand_resname}.pdb"), "w") as out_lig:
                out_lig.write(ligand_block)
                print("Ligand PDB saved in {}.".format(os.path.join(out_path, 
                                                       f"{self._ligand_resname}.pdb")))

def run(input_pdb, residue_chain, residue_number, out_pdb, ligand_resname='UNK', ligand_chain=' ', 
        verbose=False, extract_ligand=True):
    """
    It process PDB files putting the ligand part of a covalent ligand into the residue that it is attached with.

    Parameters
        ----------
        input_pdb : str
            path to input PDB file
        residue_chain : str
            chain of the residue that has the ligand covalently attached
        residue_number : int
            number of the residue that has the ligand covalently attached
        out_pdb : str
            path of the output PDB file
        ligand_resname : str
            residue name of the ligand that is covalently attached to the previous residue
        ligand_chain : str
            chain of the ligand that is covalently attached to the previous residue
        verbose : bool
            if true it will activate all the prints
        extract_ligand : bool
            if true it extracts the ligand in a separated PDB file
    """
    corrector = CovCorrector(input_pdb=input_pdb, residue_chain=residue_chain, residue_number=residue_number,
                             ligand_resname=ligand_resname, ligand_chain=ligand_chain, verbose=verbose)
    corrector.correct()
    corrector.write_file(output_file=out_pdb, extract_ligand=extract_ligand)


if __name__ == '__main__':
    input_pdb, residue_chain, residue_number, out, ligand_resname, ligand_chain, \
    verbose, extract_ligand = parseArguments()
    run(input_pdb, residue_chain, residue_number, out, ligand_resname, ligand_chain,
        verbose, extract_ligand)






import sys
import os
import re
import numpy as np
import joblib
import mdtraj
import glob
import argparse
import pandas as pd
import matplotlib.pyplot as plt


def parse_arguments():
    """
            Parse user arguments

            Output: list with all the user arguments
        """
    # All the docstrings are very provisional and some of them are old, they would be changed in further steps!!
    parser = argparse.ArgumentParser(description="""""")
    required_named = parser.add_argument_group('required named arguments')
    # Growing related arguments
    required_named.add_argument("-pdbr", "--pdb_reference",required=True,
                                help="""Reference pdb file.""")
    required_named.add_argument("-p", "--fol_path", required=True,
                                help="""Path to folder to be analyzed. """)

    parser.add_argument("-r", "--resname", default="GRW",
                        help="Resame of the ligand.")
    parser.add_argument("-if", "--index_relation_file", default=None,
                        help="File containting atom-pairing indexing to"
                             " correct the reference structure.")
    parser.add_argument("-rp", "--report_pref", default='report_',
                        help="Prefix string to find report files")
    parser.add_argument("-np", "--processors", type=int, default=4,
                        help="Number of processors to paralelize the analysis")


    # Plot arguments
    parser.add_argument("-y", "--ycol", default="RMSD",
                        help="Column name of the report file that you want to use as X axis in the plot.")
    parser.add_argument("-op", "--plot", default=False, action="store_true",
                        help="Flag to plot the results.")

    args = parser.parse_args()

    return args.pdb_reference, args.fol_path, args.resname, \
           args.index_relation_file, args.ycol, args.plot, \
           args.report_pref, args.processors


def check_common_residues(pdb_target_prody, pdb_reference_prody):
    list_of_types_and_resnums_target= []
    list_of_types_and_resnums_reference = []
    for resnum, resname in zip(pdb_target_prody.getResnums(), pdb_target_prody.getResnames()):
        list_of_types_and_resnums_target.append((resnum, resname))
    for resnum, resname in zip(pdb_reference_prody.getResnums(), pdb_reference_prody.getResnames()):
        list_of_types_and_resnums_reference.append((resnum, resname))
    common_residues = sorted(list(set(list_of_types_and_resnums_target).intersection(list_of_types_and_resnums_reference)))
    only_resnums = []
    for residue in common_residues:
        resnum = residue[0]
        only_resnums.append(resnum)
    return only_resnums


def compute_rmsd(V, W):
    """
    Calculate Root-mean-square deviation from two sets of vectors V and W.
    Parameters
    ----------
    V : array
        (N,D) matrix, where N is points and D is dimension.
    W : array
        (N,D) matrix, where N is points and D is dimension.
    Returns
    -------
    rmsd : float
        Root-mean-square deviation between the two vectors
    """
    diff = np.array(V) - np.array(W)
    N = len(V)
    return np.sqrt((diff * diff).sum() / N)


def ligands_rmsd_calculator(pdb_target, pdb_reference, resname="GRW", index_relation=None):
    """

    :param pdb_target: problem pdb file
    :param pdb_reference: reference pdb file
    :param write2report: if True export results in a file
    :param write2pdb: pdb file with the result of the superposition between "pdb_target" and "pdb_reference"
    :param ligand_chain: name of the chain of the ligand
    :return: superpose the backbone of the pdb_target to the pdb_reference and computes the RMSD of the ligand
    """
    # Load the data to mdtraj
    target = mdtraj.load(pdb_target)
    reference = mdtraj.load(pdb_reference)
    # Get the indexes of heavy atoms of the ligands
    ligand = reference.topology.select('resname "{}" and symbol != H'.format(resname))
    ligand_target = target.topology.select('resname "{}" and symbol != H'.format(resname))
    print("INDEXES REFERENCE")
    print(ligand)
    print("INDEXES TARGET")
    print(ligand_target)
    ligand_traj_ref = reference.atom_slice(ligand)
    ligand_traj_target = target.atom_slice(ligand_target)
    ref_coords = ligand_traj_ref.xyz
    target_coords = ligand_traj_target.xyz
    if index_relation:
        with open(index_relation, "r") as in_f:
            indexes = in_f.readlines()
        list_of_pairs = []
        indexes_target = []
        indexes_ref = []
        # Read
        for idx_pair in indexes:
            idx_ref, idx_tar = idx_pair.split()
            list_of_pairs.append([int(idx_ref), int(idx_tar)])
            indexes_target.append(int(idx_tar))
            indexes_ref.append(int(idx_ref))
        # Get ordering for indexes from the target
        index_target_order = {}
        for n, index in enumerate(sorted(indexes_target)):
            index_target_order[index] = n
        # Get ordering for indexes from the reference
        index_ref_order = {}
        for n, index in enumerate(sorted(indexes_ref)):
            index_ref_order[index] = n
        new_coords = {}
        print("NEW LIGAND TARGET INDEXING RELATION:")
        for pair in sorted(list_of_pairs):
            idx_ref, idx_tar = pair
            print(index_ref_order[idx_ref], index_target_order[idx_tar])
            print(ligand_traj_ref.topology.atom(index_ref_order[idx_ref]), 
                  ligand_traj_target.topology.atom(index_target_order[idx_tar]))
            new_coords[index_target_order[idx_tar]] = ref_coords[0][index_ref_order[idx_ref]]
        ref_coords_new = []
        for key in sorted(new_coords):
            coords = new_coords[key]
            ref_coords_new.append(coords)
        ref_coords = []
        ref_coords.append(ref_coords_new)
            
    rmsd_list = []
    for snap in target_coords:
        rmsd = compute_rmsd(ref_coords[0], snap) * 10 # from nm to A
        rmsd_list.append(rmsd)
    return rmsd_list


def repair_pdbs(pdb_file):
    with open(pdb_file) as pdb:
        pdb_content = pdb.readlines()
    pdb_new = []
    for line in pdb_content:
        atom_name = line[12:16]
        if line.startswith("HETATM") and atom_name.strip().startswith("G"):
            line_new = list(line)
            size = len(atom_name.strip())
            white_spaces = 3-size
            line_new[12:16] = line[77:78] + atom_name.strip() + " "*white_spaces
            modified = "".join(line_new)
            pdb_new.append(modified)
        else:
            pdb_new.append(line)
    pdb_modified = "".join(pdb_new)
    with open(pdb_file, "w") as write_output:
        write_output.write(pdb_modified)


def sort_atoms_by_name(pdb_file):
    with open(pdb_file) as pdb:
        pdb_content = pdb.readlines()
    pdb_content_sorted = (sorted(pdb_content, key=lambda atom_name: atom_name[12:16]))
    new_pdb_content = "".join(pdb_content_sorted)
    with open(pdb_file, "w") as write_output:
        write_output.write(new_pdb_content)

def read_reports(path_to_folder, report_pref):
    """
        This function merge the content of different report for PELE simulations in a single file pandas Data Frame.
    """
    data = []
    report_list = glob.glob('{}[0-9]*'.format(os.path.join(path_to_folder, report_pref)))
    for report in report_list:
        tmp_data = pd.read_csv(report, sep='    ', engine='python')
        tmp_data = tmp_data.iloc[1:]  # We must discard the first row
        processor = re.findall('\d+$'.format(path_to_folder), report)
        tmp_data['Processor'] = processor[0]
        data.append(tmp_data)
    result = pd.concat(data)
    return result

def compute_trajectory_rmsd(traj, fol_path, pdb_reference, resname, 
                            index_relation, report_pref):
    # Compute RMSD
    result = ligands_rmsd_calculator(traj, pdb_reference, resname=resname,
                                     index_relation=index_relation)
    processor = re.findall(r'\d+', traj.split("/")[-1])[0]
    # Store data in reports
    report = "{}".format(os.path.join(fol_path,
                         "{}".format(report_pref + processor)))
    print(report, traj)
    try:
        new_lines = []
        with open(report) as rep:
            rep_lines = rep.readlines()
            rep_lines = [x.strip("\n") for x in rep_lines]
        for ind, line in enumerate(rep_lines):
            new_content = list(line.split("    "))
 
            if new_content[-1] == '':
                new_content = new_content[:-1]
 
            if ind == 0:
                new_content.append('rmsd')
            else:
                value = "{:.3f}".format(result[ind-1])
                new_content.append(value)
            new_line = "    ".join(new_content)
            new_lines.append(new_line)
        new_report = "\n".join(new_lines)
        new_rep_name = report.split("/")
        new_rep_name[-1] = "rmsd_" + new_rep_name[-1]
        new_rep_name = "/".join(new_rep_name)
        with open(new_rep_name, "w") as out:
            out.write(new_report)
    except Exception as e:
        print(f"{e} error in {report}!")

def compute_rmsd_in_serie(pdb_reference, fol_path, resname="GRW", 
                          plot=False, y=None, index_relation=None,
                          report_pref='report_', processors=4):
    equilibration_files = sorted(glob.glob("{}/trajectory_[0-9]*.pdb".format(fol_path)))
    # Compute RMSD
    joblib.Parallel(n_jobs=processors)(joblib.delayed(compute_trajectory_rmsd)
                   (traj, fol_path, pdb_reference, resname, index_relation, report_pref) 
                   for traj in equilibration_files)
    # Store a CSV
    df = read_reports(fol_path, "rmsd_{}".format(report_pref))
    # Overwriting the initial report
    df.to_csv(os.path.join(fol_path, 'rmsd_summary.csv'))
    if plot:
        filename = os.path.join(fol_path, "{}_rmsd.png".format(y))
        plot_results(df, xcolumn="rmsd", ycolumn=y, outfile=filename)


def check_if_multimodels_and_split(pdb_file):
    with open(pdb_file) as pdb:
        content = pdb.readlines()
    pdbs = []
    lines_in_model = []
    for line in content:
        lines_in_model.append(line)
        if "ENDMDL" in line:
            pdbs.append("".join(lines_in_model))
            lines_in_model = []
    return pdbs


def get_report_file_from_trajectory(trajectory_path, report_prefix):
    regex = re.compile(r'\d+')
    sufix = regex.findall(trajectory_path)[-1]
    prefix = trajectory_path.split("/")[0:-1]
    new_prefix = "/".join(prefix)
    path_completed = os.path.join(new_prefix, report_prefix+sufix)
    return path_completed


def plot_results(dataframe, xcolumn, ycolumn, outfile):
    dataframe.plot(xcolumn, ycolumn, kind="scatter")
    plt.xlabel(xcolumn)
    plt.ylabel(ycolumn)
    plt.savefig(outfile)
    plt.show()


def get_snapshot(trajectory_file, model, output_file):
    pdb = check_if_multimodels_and_split(trajectory_file)[model]
    with open(output_file,"w") as pdb2write:
        pdb2write.write(pdb)


def get_min_row(data, column_number):
    minimum = data.loc[data[column_number] == data[column_number].min()]
    return minimum


def get_pdb_by_min_value(report_file, column_number):
    data = pd.read_csv(report_file, sep="\t", header=None)
    minium_row = get_min_row(data, column_number)
    return minium_row[0].values[0], minium_row[1].values[0]


if __name__ == '__main__':
    pdb_reference, fol_path, resname, indx_rel, ycol, plot, \
    report_pref, processors = parse_arguments()
    compute_rmsd_in_serie(pdb_reference, fol_path, resname=resname, 
                          plot=plot, y=ycol, index_relation=indx_rel, 
                          report_pref=report_pref, processors=processors)

